@@ -10,11 +10,11 @@
  *
  * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU Library General Public License for more details.
  *
- * You should have received a copy of the GNU Library General 
- * Public License along with this library; if not, write to the 
+ * You should have received a copy of the GNU Library General
+ * Public License along with this library; if not, write to the
  * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
  * Boston, MA 02110-1301, USA.
  *
@@ -22,9 +22,12 @@
  * for libgnomeui.
  */
 
-#ifdef HAVE_CONFIG_H
-#include <config.h>
-#endif
+#include "jpeg-thumbnailer.h"
+
+#include <gdk-pixbuf/gdk-pixbuf.h>
+#include <glib-object.h>
+#include <glib/gi18n.h>
+#include <jpeglib.h>
 
 #ifdef HAVE_SYS_MMAN_H
 #include <sys/mman.h>
@@ -35,7 +38,6 @@
 #ifdef HAVE_SYS_TYPES_H
 #include <sys/types.h>
 #endif
-
 #ifdef HAVE_FCNTL_H
 #include <fcntl.h>
 #endif
@@ -58,34 +60,20 @@
 #include <unistd.h>
 #endif
 
-#include <math.h>
-
-#include <jpeglib.h>
-
-#include <glib.h>
-#include <glib/gi18n.h>
-#include <glib-object.h>
-
-#include <gdk-pixbuf/gdk-pixbuf.h>
-
-#include <tumbler/tumbler.h>
-
-#include <jpeg-thumbnailer/jpeg-thumbnailer.h>
 
 
-
-static void fatal_error_handler     (j_common_ptr                cinfo) G_GNUC_NORETURN;
-static void jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
-                                     GCancellable               *cancellable,
-                                     TumblerFileInfo            *info);
+static void
+fatal_error_handler (j_common_ptr cinfo) G_GNUC_NORETURN;
+static void
+tvtj_free (guchar *pixels,
+           gpointer data);
+static void
+jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
+                         GCancellable *cancellable,
+                         TumblerFileInfo *info);
 
 
 
-struct _JPEGThumbnailerClass
-{
-  TumblerAbstractThumbnailerClass __parent__;
-};
-
 struct _JPEGThumbnailer
 {
   TumblerAbstractThumbnailer __parent__;
@@ -93,7 +81,7 @@ struct _JPEGThumbnailer
 
 
 
-G_DEFINE_DYNAMIC_TYPE (JPEGThumbnailer, 
+G_DEFINE_DYNAMIC_TYPE (JPEGThumbnailer,
                        jpeg_thumbnailer,
                        TUMBLER_TYPE_ABSTRACT_THUMBNAILER);
 
@@ -142,7 +130,7 @@ tvtj_noop (void)
 typedef struct
 {
   struct jpeg_error_mgr mgr;
-  jmp_buf               setjmp_buffer;
+  jmp_buf setjmp_buffer;
 } TvtjErrorHandler;
 
 
@@ -156,6 +144,15 @@ fatal_error_handler (j_common_ptr cinfo)
 
 
 
+static void
+tvtj_free (guchar *pixels,
+           gpointer data)
+{
+  g_free (pixels);
+}
+
+
+
 static gboolean
 tvtj_fill_input_buffer (j_decompress_ptr cinfo)
 {
@@ -164,8 +161,7 @@ tvtj_fill_input_buffer (j_decompress_ptr cinfo)
   /* return a fake EOI marker so we will eventually terminate */
   if (G_LIKELY (source->bytes_in_buffer == 0))
     {
-      static const JOCTET FAKE_EOI[2] =
-      {
+      static const JOCTET FAKE_EOI[2] = {
         (JOCTET) 0xff,
         (JOCTET) JPEG_EOI,
       };
@@ -181,7 +177,7 @@ tvtj_fill_input_buffer (j_decompress_ptr cinfo)
 
 static void
 tvtj_skip_input_data (j_decompress_ptr cinfo,
-                      glong            num_bytes)
+                      glong num_bytes)
 {
   struct jpeg_source_mgr *source = cinfo->src;
 
@@ -215,10 +211,10 @@ tvtj_denom (gint width,
 
 static inline void
 tvtj_convert_cmyk_to_rgb (j_decompress_ptr cinfo,
-                          guchar          *line)
+                          guchar *line)
 {
   guchar *p;
-  gint    c, k, m, n, y;
+  gint c, k, m, n, y;
 
   g_return_if_fail (cinfo != NULL);
   g_return_if_fail (cinfo->output_components == 4);
@@ -250,20 +246,20 @@ tvtj_convert_cmyk_to_rgb (j_decompress_ptr cinfo,
 
 
 
-static GdkPixbuf*
+static GdkPixbuf *
 tvtj_jpeg_load (const JOCTET *content,
-                gsize         length,
-                gint          size)
+                gsize length,
+                gint size)
 {
   struct jpeg_decompress_struct cinfo;
-  struct jpeg_source_mgr        source;
-  TvtjErrorHandler              handler;
-  guchar                       *lines[1];
-  guchar                       *buffer = NULL;
-  guchar                       *pixels = NULL;
-  guchar                       *p;
-  gint                          out_num_components;
-  guint                         n;
+  struct jpeg_source_mgr source;
+  TvtjErrorHandler handler;
+  guchar *lines[1];
+  guchar *buffer = NULL;
+  guchar *pixels = NULL;
+  guchar *p;
+  gint out_num_components;
+  guint n;
 
   /* setup JPEG error handling */
   cinfo.err = jpeg_std_error (&handler.mgr);
@@ -313,6 +309,7 @@ tvtj_jpeg_load (const JOCTET *content,
   if (G_LIKELY (cinfo.num_components != 1))
     {
       pixels = g_malloc (cinfo.output_width * cinfo.output_height * cinfo.num_components);
+      buffer = NULL;
       out_num_components = cinfo.num_components;
       lines[0] = pixels;
     }
@@ -325,12 +322,12 @@ tvtj_jpeg_load (const JOCTET *content,
     }
 
   /* process the JPEG data */
-  for (p = pixels; cinfo.output_scanline < cinfo.output_height; )
+  for (p = pixels; cinfo.output_scanline < cinfo.output_height;)
     {
       jpeg_read_scanlines (&cinfo, lines, 1);
 
       /* convert the data to RGB */
-      if (cinfo.num_components == 1)
+      if (cinfo.num_components == 1 && buffer != NULL)
         {
           for (n = 0; n < cinfo.output_width; ++n)
             {
@@ -361,7 +358,7 @@ tvtj_jpeg_load (const JOCTET *content,
                                    (cinfo.out_color_components == 4), 8,
                                    cinfo.output_width, cinfo.output_height,
                                    cinfo.output_width * out_num_components,
-                                   (GdkPixbufDestroyNotify) g_free, NULL);
+                                   tvtj_free, NULL);
 
 error:
   jpeg_destroy_decompress (&cinfo);
@@ -375,34 +372,33 @@ error:
 typedef struct
 {
   const guchar *data_ptr;
-  guint         data_len;
+  guint data_len;
 
-  guint         thumb_compression;
-  union
+  guint thumb_compression;
+
+  struct /* thumbnail JPEG */
   {
-    struct /* thumbnail JPEG */
-    {
-      guint     length;
-      guint     offset;
-    } thumb_jpeg;
-    struct /* thumbnail TIFF */
-    {
-      guint     length;
-      guint     offset;
-      guint     interp;
-      guint     height;
-      guint     width;
-    } thumb_tiff;
-  } thumb;
+    guint length;
+    guint offset;
+    guint orientation;
+  } thumb_jpeg;
+  struct /* thumbnail TIFF */
+  {
+    guint length;
+    guint offset;
+    guint interp;
+    guint height;
+    guint width;
+  } thumb_tiff;
 
-  gboolean      big_endian;
+  gboolean big_endian;
 } TvtjExif;
 
 
 
 static guint
 tvtj_exif_get_ushort (const TvtjExif *exif,
-                      const guchar   *data)
+                      const guchar *data)
 {
   if (G_UNLIKELY (exif->big_endian))
     return ((data[0] << 8) | data[1]);
@@ -414,7 +410,7 @@ tvtj_exif_get_ushort (const TvtjExif *exif,
 
 static guint
 tvtj_exif_get_ulong (const TvtjExif *exif,
-                     const guchar   *data)
+                     const guchar *data)
 {
   if (G_UNLIKELY (exif->big_endian))
     return ((data[0] << 24) | (data[1] << 16) | (data[2] << 8) | data[3]);
@@ -425,17 +421,17 @@ tvtj_exif_get_ulong (const TvtjExif *exif,
 
 
 static void
-tvtj_exif_parse_ifd (TvtjExif     *exif,
+tvtj_exif_parse_ifd (TvtjExif *exif,
                      const guchar *ifd_ptr,
-                     guint         ifd_len,
-                     GSList       *ifd_previous_list)
+                     guint ifd_len,
+                     GSList *ifd_previous_list)
 {
   const guchar *subifd_ptr;
-  GSList        ifd_list;
-  guint         subifd_off;
-  guint         value;
-  guint         tag;
-  guint         n;
+  GSList ifd_list;
+  guint subifd_off;
+  guint value;
+  guint tag;
+  guint n;
 
   /* make sure we have a valid IFD here */
   if (G_UNLIKELY (ifd_len < 2))
@@ -495,15 +491,15 @@ tvtj_exif_parse_ifd (TvtjExif     *exif,
 
           /* and remember it appropriately */
           if (tag == 0x0100)
-            exif->thumb.thumb_tiff.width = value;
-          else if (tag == 0x0100)
-            exif->thumb.thumb_tiff.height = value;
+            exif->thumb_tiff.width = value;
+          else if (tag == 0x0101)
+            exif->thumb_tiff.height = value;
           else if (tag == 0x0106)
-            exif->thumb.thumb_tiff.interp = value;
+            exif->thumb_tiff.interp = value;
           else if (tag == 0x0111)
-            exif->thumb.thumb_tiff.offset = value;
+            exif->thumb_tiff.offset = value;
           else
-            exif->thumb.thumb_tiff.length = value;
+            exif->thumb_tiff.length = value;
         }
       else if (tag == 0x0201 || tag == 0x0202)
         {
@@ -515,9 +511,20 @@ tvtj_exif_parse_ifd (TvtjExif     *exif,
 
               /* and remember it appropriately */
               if (G_LIKELY (tag == 0x201))
-                exif->thumb.thumb_jpeg.offset = value;
+                exif->thumb_jpeg.offset = value;
               else
-                exif->thumb.thumb_jpeg.length = value;
+                exif->thumb_jpeg.length = value;
+            }
+        }
+      else if (tag == 0x112)
+        {
+          /* verify that we have a 2-byte integer (format 3) */
+          if (tvtj_exif_get_ushort (exif, ifd_ptr + 2) == 3
+              && tvtj_exif_get_ulong (exif, ifd_ptr + 4) == 1)
+            {
+              /* determine the orientation value */
+              value = tvtj_exif_get_ushort (exif, ifd_ptr + 8);
+              exif->thumb_jpeg.orientation = MIN (value, 8);
             }
         }
     }
@@ -533,13 +540,74 @@ tvtj_exif_parse_ifd (TvtjExif     *exif,
 
 
 
-static GdkPixbuf*
+static GdkPixbuf *
+tvtj_rotate_pixbuf (GdkPixbuf *src,
+                    guint orientation)
+{
+  GdkPixbuf *dest = NULL;
+  GdkPixbuf *temp;
+
+  g_return_val_if_fail (GDK_IS_PIXBUF (src), NULL);
+
+  switch (orientation)
+    {
+    case 1:
+      dest = g_object_ref (src);
+      break;
+
+    case 2:
+      dest = gdk_pixbuf_flip (src, TRUE);
+      break;
+
+    case 3:
+      dest = gdk_pixbuf_rotate_simple (src, GDK_PIXBUF_ROTATE_UPSIDEDOWN);
+      break;
+
+    case 4:
+      dest = gdk_pixbuf_flip (src, FALSE);
+      break;
+
+    case 5:
+      temp = gdk_pixbuf_rotate_simple (src, GDK_PIXBUF_ROTATE_CLOCKWISE);
+      dest = gdk_pixbuf_flip (temp, TRUE);
+      g_object_unref (temp);
+      break;
+
+    case 6:
+      dest = gdk_pixbuf_rotate_simple (src, GDK_PIXBUF_ROTATE_CLOCKWISE);
+      break;
+
+    case 7:
+      temp = gdk_pixbuf_rotate_simple (src, GDK_PIXBUF_ROTATE_CLOCKWISE);
+      dest = gdk_pixbuf_flip (temp, FALSE);
+      g_object_unref (temp);
+      break;
+
+    case 8:
+      dest = gdk_pixbuf_rotate_simple (src, GDK_PIXBUF_ROTATE_COUNTERCLOCKWISE);
+      break;
+
+    default:
+      /* if no orientation tag was present */
+      dest = g_object_ref (src);
+      break;
+    }
+
+  return dest;
+}
+
+
+
+static GdkPixbuf *
 tvtj_exif_extract_thumbnail (const guchar *data,
-                             guint         length,
-                             gint          size)
+                             guint length,
+                             gint size,
+                             guint *exif_orientation)
 {
   TvtjExif exif;
-  guint    offset;
+  guint offset;
+  GdkPixbuf *thumb = NULL;
+  GdkPixbuf *rotated;
 
   /* make sure we have enough data */
   if (G_UNLIKELY (length < 6 + 8))
@@ -583,40 +651,54 @@ tvtj_exif_extract_thumbnail (const guchar *data,
       if (G_LIKELY (exif.thumb_compression == 6)) /* JPEG */
         {
           /* check if we have a valid thumbnail JPEG */
-          if (exif.thumb.thumb_jpeg.offset > 0 && exif.thumb.thumb_jpeg.length > 0
-              && exif.thumb.thumb_jpeg.offset + exif.thumb.thumb_jpeg.length <= length)
+          if (exif.thumb_jpeg.offset > 0 && exif.thumb_jpeg.length > 0
+              && exif.thumb_jpeg.offset + exif.thumb_jpeg.length <= length)
             {
               /* try to load the embedded thumbnail JPEG */
-              return tvtj_jpeg_load (data + exif.thumb.thumb_jpeg.offset, exif.thumb.thumb_jpeg.length, size);
+              thumb = tvtj_jpeg_load (data + exif.thumb_jpeg.offset, exif.thumb_jpeg.length, size);
             }
         }
       else if (exif.thumb_compression == 1) /* Uncompressed */
         {
           /* check if we have a valid thumbnail (current only RGB interpretations) */
-          if (G_LIKELY (exif.thumb.thumb_tiff.interp == 2)
-              && exif.thumb.thumb_tiff.offset > 0 && exif.thumb.thumb_tiff.length > 0
-              && exif.thumb.thumb_tiff.offset + exif.thumb.thumb_tiff.length <= length
-              && exif.thumb.thumb_tiff.height * exif.thumb.thumb_tiff.width == exif.thumb.thumb_tiff.length)
+          if (G_LIKELY (exif.thumb_tiff.interp == 2)
+              && exif.thumb_tiff.offset > 0 && exif.thumb_tiff.length > 0
+              && exif.thumb_tiff.offset + exif.thumb_tiff.length <= length
+              && exif.thumb_tiff.height * exif.thumb_tiff.width == exif.thumb_tiff.length)
             {
               /* plain RGB data, just what we need for a GdkPixbuf */
-              return gdk_pixbuf_new_from_data (g_memdup (data + exif.thumb.thumb_tiff.offset, exif.thumb.thumb_tiff.length),
-                                               GDK_COLORSPACE_RGB, FALSE, 8, exif.thumb.thumb_tiff.width,
-                                               exif.thumb.thumb_tiff.height, exif.thumb.thumb_tiff.width,
-                                               (GdkPixbufDestroyNotify) g_free, NULL);
+              thumb = gdk_pixbuf_new_from_data (g_memdup2 (data + exif.thumb_tiff.offset, exif.thumb_tiff.length),
+                                                GDK_COLORSPACE_RGB, FALSE, 8, exif.thumb_tiff.width,
+                                                exif.thumb_tiff.height, exif.thumb_tiff.width,
+                                                tvtj_free, NULL);
             }
         }
+
+      *exif_orientation = exif.thumb_jpeg.orientation;
+
+      if (thumb != NULL
+          && exif.thumb_jpeg.orientation > 1)
+        {
+          /* rotate thumbnail */
+          rotated = tvtj_rotate_pixbuf (thumb, exif.thumb_jpeg.orientation);
+          g_object_unref (thumb);
+          thumb = rotated;
+        }
     }
 
-  return NULL;
+  return thumb;
 }
 
 
 
-static GdkPixbuf*
+static GdkPixbuf *
 tvtj_jpeg_load_thumbnail (const JOCTET *content,
-                          gsize         length,
-                          gint          size)
+                          gsize length,
+                          gint width,
+                          gint height,
+                          guint *exif_orientation)
 {
+  GdkPixbuf *pixbuf = NULL;
   guint marker_len;
   guint marker;
   gsize n;
@@ -625,7 +707,7 @@ tvtj_jpeg_load_thumbnail (const JOCTET *content,
   if (G_LIKELY (length >= 2 && content[0] == 0xff && content[1] == 0xd8))
     {
       /* search for an EXIF marker */
-      for (length -= 2, n = 2; n < length; )
+      for (length -= 2, n = 2; n < length;)
         {
           /* check for valid marker start */
           if (G_UNLIKELY (content[n++] != 0xff))
@@ -656,7 +738,15 @@ tvtj_jpeg_load_thumbnail (const JOCTET *content,
           if (marker == 0xe1 && n + marker_len <= length)
             {
               /* try to extract the Exif thumbnail */
-              return tvtj_exif_extract_thumbnail (content + n + 2, marker_len - 2, size);
+              pixbuf = tvtj_exif_extract_thumbnail (content + n + 2, marker_len - 2,
+                                                    MIN (width, height), exif_orientation);
+
+              /* do not use low quality embedded thumbnail */
+              if (pixbuf != NULL && gdk_pixbuf_get_width (pixbuf) < width
+                  && gdk_pixbuf_get_height (pixbuf) < height)
+                g_clear_object (&pixbuf);
+              else
+                break;
             }
 
           /* try next one then */
@@ -664,79 +754,43 @@ tvtj_jpeg_load_thumbnail (const JOCTET *content,
         }
     }
 
-  return NULL;
-}
-
-
-
-static GdkPixbuf*
-scale_pixbuf (GdkPixbuf *source,
-              gint       dest_width,
-              gint       dest_height)
-{
-  gdouble wratio;
-  gdouble hratio;
-  gint    source_width;
-  gint    source_height;
-
-  /* determine source pixbuf dimensions */
-  source_width  = gdk_pixbuf_get_width  (source);
-  source_height = gdk_pixbuf_get_height (source);
-
-  /* don't do anything if there is no need to resize */
-  if (source_width <= dest_width && source_height <= dest_height)
-    return g_object_ref (source);
-
-  /* determine which axis needs to be scaled down more */
-  wratio = (gdouble) source_width  / (gdouble) dest_width;
-  hratio = (gdouble) source_height / (gdouble) dest_height;
-
-  /* adjust the other axis */
-  if (hratio > wratio)
-    dest_width = rint (source_width / hratio);
-  else
-    dest_height = rint (source_height / wratio);
-
-  /* scale the pixbuf down to the desired size */
-  return gdk_pixbuf_scale_simple (source, MAX (dest_width, 1),
-                                  MAX (dest_height, 1),
-                                  GDK_INTERP_BILINEAR);
+  return pixbuf;
 }
 
 
 
 static void
 jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
-                         GCancellable               *cancellable,
-                         TumblerFileInfo            *info)
+                         GCancellable *cancellable,
+                         TumblerFileInfo *info)
 {
   TumblerThumbnailFlavor *flavor;
-  TumblerImageData        data;
-  TumblerThumbnail       *thumbnail;
-  struct stat             statb;
-  const gchar            *uri;
-  GdkPixbuf              *pixbuf = NULL;
-  GdkPixbuf              *scaled;
-  gboolean                streaming_needed = TRUE;
-  JOCTET                 *content;
-  GError                 *error = NULL;
-  GFile                  *file;
-  gchar                  *path;
-  gsize                   length;
-  gint                    fd;
-  gint                    height;
-  gint                    width;
-  gint                    size;
+  TumblerImageData data;
+  TumblerThumbnail *thumbnail;
+  struct stat statb;
+  const gchar *uri;
+  GdkPixbuf *pixbuf = NULL;
+  GdkPixbuf *scaled;
+  gboolean streaming_needed = TRUE;
+  JOCTET *content;
+  GError *error = NULL;
+  GFile *file;
+  gsize length;
+  gint fd;
+  gint height;
+  gint width;
+  gint size;
 
-  g_return_if_fail (IS_JPEG_THUMBNAILER (thumbnailer));
+  g_return_if_fail (JPEG_IS_THUMBNAILER (thumbnailer));
   g_return_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable));
   g_return_if_fail (TUMBLER_IS_FILE_INFO (info));
 
   /* do nothing if cancelled */
-  if (g_cancellable_is_cancelled (cancellable)) 
+  if (g_cancellable_is_cancelled (cancellable))
     return;
 
   uri = tumbler_file_info_get_uri (info);
+  g_debug ("Handling URI '%s'", uri);
 
   /* try to open the source file for reading */
   file = g_file_new_for_uri (uri);
@@ -751,36 +805,43 @@ jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
   size = MIN (width, height);
 
 #ifdef HAVE_MMAP
-  if (g_file_is_native (file))
+  if (g_file_peek_path (file) != NULL)
     {
-      path = g_file_get_path (file);
-
       /* try to open the file at the given path */
-      fd = open (path, O_RDONLY);
+      fd = open (g_file_peek_path (file), O_RDONLY);
       if (G_LIKELY (fd >= 0))
         {
           /* determine the status of the file */
           if (G_LIKELY (fstat (fd, &statb) == 0 && statb.st_size > 0))
             {
               /* try to mmap the file */
-              content = (JOCTET *) mmap (NULL, statb.st_size, PROT_READ, 
+              content = (JOCTET *) mmap (NULL, statb.st_size, PROT_READ,
                                          MAP_SHARED, fd, 0);
 
               /* verify whether the mmap was successful */
               if (G_LIKELY (content != (JOCTET *) MAP_FAILED))
                 {
+                  guint exif_orientation = 0;
+
                   /* try to load the embedded thumbnail first */
-                  pixbuf = tvtj_jpeg_load_thumbnail (content, statb.st_size, size);
+                  pixbuf = tvtj_jpeg_load_thumbnail (content, statb.st_size, width, height,
+                                                     &exif_orientation);
                   if (pixbuf == NULL)
                     {
                       /* fall back to loading and scaling the image itself */
                       pixbuf = tvtj_jpeg_load (content, statb.st_size, size);
 
-                      if (pixbuf == NULL)
+                      if (G_UNLIKELY (pixbuf == NULL))
                         {
-                          g_set_error (&error, TUMBLER_ERROR, 
-                                       TUMBLER_ERROR_INVALID_FORMAT,
-                                       _("Thumbnail could not be inferred from file contents"));
+                          g_set_error (&error, TUMBLER_ERROR, TUMBLER_ERROR_INVALID_FORMAT,
+                                       TUMBLER_ERROR_MESSAGE_CREATION_FAILED);
+                        }
+                      else if (exif_orientation > 1)
+                        {
+                          GdkPixbuf *rotated;
+                          rotated = tvtj_rotate_pixbuf (pixbuf, exif_orientation);
+                          g_object_unref (pixbuf);
+                          pixbuf = rotated;
                         }
                     }
 
@@ -798,27 +859,33 @@ jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
           /* close the file */
           close (fd);
         }
-
-      g_free (path);
     }
 #endif
 
   if (streaming_needed)
     {
-      g_file_load_contents (file, cancellable, (gchar **)&content, &length, 
+      g_file_load_contents (file, cancellable, (gchar **) &content, &length,
                             NULL, &error);
 
       if (error == NULL)
         {
-          pixbuf = tvtj_jpeg_load_thumbnail (content, length, size);
+          guint exif_orientation = 0;
+          pixbuf = tvtj_jpeg_load_thumbnail (content, length, width, height, &exif_orientation);
 
           if (pixbuf == NULL)
             {
               pixbuf = tvtj_jpeg_load (content, length, size);
-              if (pixbuf == NULL)
+              if (G_UNLIKELY (pixbuf == NULL))
                 {
                   g_set_error (&error, TUMBLER_ERROR, TUMBLER_ERROR_INVALID_FORMAT,
-                               _("Thumbnail could not be inferred from file contents"));
+                               TUMBLER_ERROR_MESSAGE_CREATION_FAILED);
+                }
+              else if (exif_orientation > 1)
+                {
+                  GdkPixbuf *rotated;
+                  rotated = tvtj_rotate_pixbuf (pixbuf, exif_orientation);
+                  g_object_unref (pixbuf);
+                  pixbuf = rotated;
                 }
             }
         }
@@ -829,7 +896,7 @@ jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
 
   if (pixbuf != NULL)
     {
-      scaled = scale_pixbuf (pixbuf, width, height);
+      scaled = tumbler_util_scale_pixbuf (pixbuf, width, height);
       g_object_unref (pixbuf);
       pixbuf = scaled;
 
@@ -841,8 +908,8 @@ jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
       data.rowstride = gdk_pixbuf_get_rowstride (pixbuf);
       data.colorspace = (TumblerColorspace) gdk_pixbuf_get_colorspace (pixbuf);
 
-      tumbler_thumbnail_save_image_data (thumbnail, &data, 
-                                         tumbler_file_info_get_mtime (info), 
+      tumbler_thumbnail_save_image_data (thumbnail, &data,
+                                         tumbler_file_info_get_mtime (info),
                                          NULL, &error);
 
       g_object_unref (pixbuf);
@@ -850,12 +917,13 @@ jpeg_thumbnailer_create (TumblerAbstractThumbnailer *thumbnailer,
 
   if (error != NULL)
     {
-      g_signal_emit_by_name (thumbnailer, "error", uri, error->code, error->message);
+      g_signal_emit_by_name (thumbnailer, "error", info,
+                             error->domain, error->code, error->message);
       g_error_free (error);
     }
   else
     {
-      g_signal_emit_by_name (thumbnailer, "ready", uri);
+      g_signal_emit_by_name (thumbnailer, "ready", info);
     }
 
   g_object_unref (flavor);
